//*****************************************************************************
//*****************************************************************************
//  FILENAME:   sensor.c
//  Version: 1.0, 2007/03/08
//  Generated by Andy Kiser
//
//  DESCRIPTION:  sensor specific functions 
//
//-----------------------------------------------------------------------------
// Originally authored by Toradex AG. This work is now available under the terms and conditions
// of the Creative Commons License 'Attribution CC BY', details of which can be found here:
// http://creativecommons.org/licenses/by/3.0/
//*****************************************************************************
//*****************************************************************************
//#define __DEBUG__
#include <m8c.h>        // part specific constants and macros
#include "sensor.h"
#include "psocgpioint.h"
#include "psocapi.h"
#include "feature.h"
#include "parameters.h"
#include "types_n_macros.h"
//----------------------------------------
// Sensor parameters
// Factory Defaults are defined here (NUM_xxx are defined in "sensor.h")
//----------------------------------------
const BYTE FactoryDefault_Param_8   [NUM_PARAM_8] =
{
	0x00,		// report mode
	0x00,		// LED mode
	0x00		// Sampling mode for the AS5045 (0=data, 1=align)
};

const WORD FactoryDefault_Param_16  [NUM_PARAM_16] = 
{
	500,		// report rate [ms]
	500,		// sample rate [ms]
	50000,		// sample rate [us]	(not used for this sensor)
	0,			// OTP register of the AS5046 (user part)
	0x3C01		// OTP register of the AS5046 (factory setting part)
};

const DWORD FactoryDefault_Param_32  [NUM_PARAM_32] =
{
	0x00000000	// currently no 32bit parameters
};

const tParam_168 FactoryDefault_Param_168 [NUM_PARAM_168] =
{
	"Rotary Encoder",	// user device name
	"Frame Number",		// user name Channel 0
	"Angle",			// user name Channel 1
	"Magnitude",		// user name Channel 2
	"Status"			// user name Channel 3
};


//----------------------------------------
// this module's global variables
//----------------------------------------
tReport Report;

//----------------------------------------
// function Prototypes
//----------------------------------------
static void Goto_I2C(void);
static void Goto_SPI(void);

//----------------------------------------
// constant definitions
//----------------------------------------
#define TYPE_MASK   0xF8
#define ADDR_MASK   0x07

#define TYPE_SIU	0x28
#define TYPE_HALLFE	0x08
#define TYPE_ADC	0x20
#define TYPE_AGC	0x38

#define ADDR_SIU		0x00
#define ADDR_HALLFE0	0x00
#define ADDR_HALLFE1	0x01
#define ADDR_HALLFE2	0x02
#define ADDR_HALLFE3	0x03
#define ADDR_HALLFE4	0x04
#define ADDR_HALLFE5	0x05
#define ADDR_HALLFE6	0x06
#define ADDR_HALLFE7	0x07
#define ADDR_ADC		0x00
#define ADDR_AGC		0x00

//----------------------------------------
// constant lookup-table for arc sin function
//----------------------------------------
const WORD Ang_MSB[16] =
{
	0,3927,7854,11781,15708,19635,23562,27489,
	31416,35343,39270,43197,47124,51051,54978,58905
};

const WORD Ang_LSB[256] =
{
	0,15,31,46,61,77,92,107,
	123,138,153,169,184,199,215,230,
	245,261,276,291,307,322,337,353,
	368,383,399,414,430,445,460,476,
	491,506,522,537,552,568,583,598,
	614,629,644,660,675,690,706,721,
	736,752,767,782,798,813,828,844,
	859,874,890,905,920,936,951,966,
	982,997,1012,1028,1043,1058,1074,1089,
	1104,1120,1135,1150,1166,1181,1197,1212,
	1227,1243,1258,1273,1289,1304,1319,1335,
	1350,1365,1381,1396,1411,1427,1442,1457,
	1473,1488,1503,1519,1534,1549,1565,1580,
	1595,1611,1626,1641,1657,1672,1687,1703,
	1718,1733,1749,1764,1779,1795,1810,1825,
	1841,1856,1871,1887,1902,1917,1933,1948,
	1963,1979,1994,2010,2025,2040,2056,2071,
	2086,2102,2117,2132,2148,2163,2178,2194,
	2209,2224,2240,2255,2270,2286,2301,2316,
	2332,2347,2362,2378,2393,2408,2424,2439,
	2454,2470,2485,2500,2516,2531,2546,2562,
	2577,2592,2608,2623,2638,2654,2669,2684,
	2700,2715,2730,2746,2761,2777,2792,2807,
	2823,2838,2853,2869,2884,2899,2915,2930,
	2945,2961,2976,2991,3007,3022,3037,3053,
	3068,3083,3099,3114,3129,3145,3160,3175,
	3191,3206,3221,3237,3252,3267,3283,3298,
	3313,3329,3344,3359,3375,3390,3405,3421,
	3436,3451,3467,3482,3497,3513,3528,3543,
	3559,3574,3590,3605,3620,3636,3651,3666,
	3682,3697,3712,3728,3743,3758,3774,3789,
	3804,3820,3835,3850,3866,3881,3896,3912
};		

//----------------------------------------
// type definitions
//----------------------------------------

typedef enum 
{	
	ePowerdown,
	ePowerup,
	eReset,
	eIdle,
	eReadADC,
} tFSM_Measure;

//----------------------------------------
// variables

tFSM_Measure	FSM;
WORD			Countdown_1ms;

BOOL			NewSample = FALSE;
BOOL			SetOTP    = FALSE;
BOOL			SetOTP32  = FALSE;

// store original drivemodes during suspend
BYTE			DM0_2, DM0_1, DM0_0;
BYTE			DM1_2, DM1_1, DM1_0;
BYTE			DM2_2, DM2_1, DM2_0;

//-----------------------------------------------------------------------------
// Initializes the Sensor 
//-----------------------------------------------------------------------------
void Sensor_Start ()
{
	WORD Result;
	BYTE ReadMode;	
	
	FSM = eReset;

	Goto_I2C();
	
	// set Slow or Fast mode
	// The mode pin will be read by the sensor ic *only on powerup*!!!
	Get_Param_8(PAR8_READMODE, VOLATILE, &ReadMode);
	if ((ReadMode & 0x01) == 0)
		GPIO_LOW(MODE)
	else
		GPIO_HIGH(MODE)

	// turn on power for the sensor
	GPIO_LOW(OFF_nON)

	// Set MOSI default value to '0' (this is overruled by the SPI hardware)
	GPIO_LOW(MOSI)
	// Set nCS to '1'
	GPIO_HIGH(nCS)

	// initialize the Cypress' I2C interface
	I2CHW_Start();
	I2CHW_EnableMstr();
	// initialize the Cypress' SPI interface
	SPIM_Start(SPIM_SPIM_MODE_1 | SPIM_SPIM_MSB_FIRST);  // mode 1 has SCK idle=High
}
//-----------------------------------------------------------------------------
// Prepare Sensor for Suspend
//-----------------------------------------------------------------------------
void Sensor_Suspend()
{
	// turn off power for the sensor
	GPIO_HIGH(OFF_nON)
	// disable I2C hardware
	I2CHW_DisableMstr();

	// store original drivemodes during  suspend
	DM0_2 = PRT0DM2;
	DM0_1 = PRT0DM1;
	DM0_0 = PRT0DM0;
	DM1_2 = PRT1DM2;
	DM1_1 = PRT1DM1;
	DM1_0 = PRT1DM0;
	DM2_2 = PRT2DM2;
	DM2_1 = PRT2DM1;
	DM2_0 = PRT2DM0;
	
	// set all I/Os to/from the sensor to High-Z Analog (=110), 
	// except the OFF_nON signal. 
	// OFF_nON is on Port0, all other signals on Port2
	PRT0DM2 |= ~OFF_nON_MASK;
	PRT0DM1 |= ~OFF_nON_MASK;
	PRT0DM0 &=  OFF_nON_MASK;
	PRT1DM2  = 0xFF;
	PRT1DM1  = 0xFF;
	PRT1DM0  = 0x00;
	PRT2DM2  = 0xFF;
	PRT2DM1  = 0xFF;
	PRT2DM0  = 0x00;

}
//-----------------------------------------------------------------------------
// Resume Sensor from Suspend mode
//-----------------------------------------------------------------------------
void Sensor_Resume()
{
	BYTE ReadMode;
	BYTE i;
		
	// restore drivemodes for GPIOs to/from the sensor
	PRT2DM2 = DM2_2;
	PRT2DM1 = DM2_1;
	PRT2DM0 = DM2_0;
	PRT1DM2 = DM1_2;
	PRT1DM1 = DM1_1;
	PRT1DM0 = DM1_0;
	PRT0DM2 = DM0_2;
	PRT0DM1 = DM0_1;
	PRT0DM0 = DM0_0;

	// set Slow or Fast mode
	// The mode pin will be read by the sensor ic *only on powerup*!!!
	Get_Param_8(PAR8_READMODE, VOLATILE, &ReadMode);
	if ((ReadMode & 0x01) == 0)
		GPIO_LOW(MODE)
	else
		GPIO_HIGH(MODE)

	DELAY_16CYCLES	// Allow MODE 2us to settle

	// turn on power for the sensor
	GPIO_LOW(OFF_nON)
	I2CHW_EnableMstr();

	FSM = eReset;
}

//-----------------------------------------------------------------------------
// Switch to I2C mode
//  MOSI  = StdCPU, Low  (fixed)
//  nCS   = StdCPU, Low  (fixed)
//  SCK   = I2C,    High (idle state)
//  SDA   = I2C,    High (idle state)
//-----------------------------------------------------------------------------
static void Goto_I2C(void)
{
	// -- MOSI
	MOSI_GlobalSelect_ADDR &= ~MOSI_MASK;  // take manual control of the MOSI pin
	GPIO_LOW(MOSI)
	
	// -- nCS
	GPIO_LOW (nCS)
	
	// -- SCK and SDA
	GPIO_DM_OPENDRAIN(I2CHWSDA)
	I2CHW_EnableMstr();
	I2CHW_SendStop();
}
//-----------------------------------------------------------------------------
// Switch to SPI mode
//  MOSI  = GlobalOut Even 6, Low  (idle state)
//  nCS   = StdCPU          , High (idle state)
//  SCK   = GlobalOut Odd 7 , High (idle state)
//  SDA   = GlobalIn  Odd 5 , High (idle state)
//-----------------------------------------------------------------------------
static void Goto_SPI(void)
{
	// -- MOSI
	MOSI_GlobalSelect_ADDR |= MOSI_MASK;  // Allow the SPI block to control the MOSI signal

	// -- nCS
	GPIO_HIGH (nCS)
	
	I2CHW_DisableMstr();

	// -- SCK
	I2CHWSCL_GlobalSelect_ADDR |= I2CHWSCL_MASK;  // Allow the SPI block to control the SCK signal

	// -- SDA
	I2CHWSDA_GlobalSelect_ADDR |= I2CHWSDA_MASK;  // take manual control of the MOSI pin
	GPIO_DM_HIGHZ(I2CHWSDA)
}
//-----------------------------------------------------------------------------
// Read 8-bit register on I2C
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		MOSI == 1
//		SCK == 1
// 		I2C active
//-----------------------------------------------------------------------------
BYTE I2C_Read_8bit(BYTE Address) // Address also includes the Type identifier
{
	BYTE fAck;
	BYTE Data;
	
	fAck = I2CHW_fSendStart(Address, I2CHW_READ);
	Data = I2CHW_bRead(I2CHW_NAKslave);
	I2CHW_SendStop();	
	
	return Data;
}

//-----------------------------------------------------------------------------
// Read 24-bit register on I2C
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		MOSI == 1
//		SCK == 1
// 		I2C active
//-----------------------------------------------------------------------------
DWORD I2C_Read_24bit(BYTE Address) // Address also includes the Type identifier
{
	BYTE fAck;
	BYTE i;
	union
	{
		BYTE b[4];
		DWORD dw;
	} Data;
	
	fAck = I2CHW_fSendStart(Address, I2CHW_READ);
	Data.b[0] = 0;
	for (i=1; i<=3; i++)
		Data.b[i] = I2CHW_bRead(i!=3); // ACK if i!=3, NAK if i==3
	I2CHW_SendStop();	
	
	return Data.dw;
}
//-----------------------------------------------------------------------------
// Read 32-bit register on I2C
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		MOSI == 1
//		SCK == 1
// 		I2C active
//-----------------------------------------------------------------------------
DWORD I2C_Read_32bit(BYTE Address) // Address also includes the Type identifier
{
	BYTE fAck;
	BYTE i;
	union
	{
		BYTE b[4];
		DWORD dw;
	} Data;
	
	fAck = I2CHW_fSendStart(Address, I2CHW_READ);
	for (i=0; i<=3; i++)
		Data.b[i] = I2CHW_bRead(i!=3); // ACK if i!=3, NAK if i==3
	I2CHW_SendStop();	
	
	return Data.dw;
}

//-----------------------------------------------------------------------------
// Write 8-bit register on I2C
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		MOSI == 1
//		SCK == 1
// 		I2C active
//-----------------------------------------------------------------------------
void I2C_Write_8bit(BYTE Address, BYTE Data) // Address also includes the Type identifier
{
	BYTE fAck;
	
	fAck = I2CHW_fSendStart(Address, I2CHW_WRITE);
	fAck = I2CHW_fWrite (Data);
	I2CHW_SendStop();	
}

//-----------------------------------------------------------------------------
// Write 24-bit register on I2C
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		MOSI == 1
//		SCK == 1
// 		I2C active
//-----------------------------------------------------------------------------
void I2C_Write_24bit(BYTE Address, DWORD Data) // Address also includes the Type identifier
{
	BYTE fAck;
	BYTE i;
	
	fAck = I2CHW_fSendStart(Address, I2CHW_WRITE);
	for (i=1; i<=3; i++)
		fAck = I2CHW_fWrite ( ((BYTE*)&Data)[i] );
	I2CHW_SendStop();	
}
//-----------------------------------------------------------------------------
// Write 32-bit register on I2C
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		MOSI == 1
//		SCK == 1
// 		I2C active
//-----------------------------------------------------------------------------
void I2C_Write_32bit(BYTE Address, DWORD Data) // Address also includes the Type identifier
{
	BYTE fAck;
	BYTE i;
	
	fAck = I2CHW_fSendStart(Address, I2CHW_WRITE);
	for (i=0; i<=3; i++)
		fAck = I2CHW_fWrite ( ((BYTE*)&Data)[i] );
	I2CHW_SendStop();	
}

//-----------------------------------------------------------------------------
// Read Serial Interface Unit (SIU) register from the sensor
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		MOSI == 0
//		SCK == 1
// 		SPI-Mode == 1
//-----------------------------------------------------------------------------
void Sensor_ReadSIU (WORD *AngleRaw, BYTE *Magnitude, BYTE *Status)
{
	BYTE i;
	union uBigEndian AngRaw;
	union uBigEndian SIU_Addr;

	union
	{
		BYTE b[3];
		DWORD dw;
	} SIU;
	
	// -- extract SIU address from OTP register
	Get_Param_16(PAR16_SENSOR_OTP, VOLATILE, &SIU_Addr.w);
	SIU_Addr.BE.Lo &= 0x1C;	// extract SIU Address bits A2..A0
	SIU_Addr.BE.Lo >>= 2;	// shift A2..A0 to the correct position
	SIU_Addr.BE.Lo |= TYPE_SIU; // Add TYPE identifier	
	
	// -- read SIU
	SIU.dw = I2C_Read_32bit(SIU_Addr.BE.Lo);
	
	// -- extract Angle
	AngRaw.BE.Hi = SIU.b[0];
	AngRaw.BE.Lo = SIU.b[1] & 0xC0;
	SIU.b[3] >>= 2;
	AngRaw.BE.Lo |= SIU.b[3]; // D1,D0 are already shifted to bits 5,4
	AngRaw.w >>= 4;
	*AngleRaw = AngRaw.w;
	
	// -- extract Status
	SIU.b[1] &= 0x3F;
	SIU.b[1] >>=1;
	*Status = SIU.b[1];
	
	// -- extract Magnitude
	*Magnitude = SIU.b[2];
}

//-----------------------------------------------------------------------------
// Following Reads of the SUI will read Alignment data, instead of raw angle
//-----------------------------------------------------------------------------
void AlignMode_Enter(void)
{
	GPIO_HIGH(nCS)

	GPIO_HIGH(MOSI)
	MOSI_GlobalSelect_ADDR &= ~MOSI_MASK;  // take manual control of the MOSI pin
	
	DELAY_16CYCLES	// 2us setup time
	GPIO_LOW(nCS)
	DELAY_16CYCLES	// 2us hold time

	GPIO_LOW(MOSI)
}

//-----------------------------------------------------------------------------
// Following Reads of the SUI will read raw angle
//-----------------------------------------------------------------------------
void AlignMode_Exit(void)
{
	GPIO_LOW(MOSI)  // should still be low since Alignmode_Enter()

	GPIO_HIGH(nCS)	
	
	DELAY_16CYCLES	// 2us setup time
	GPIO_LOW(nCS)
	DELAY_16CYCLES	// 2us hold time
	
	MOSI_GlobalSelect_ADDR &= ~MOSI_MASK;  // give control of the MOSI pin back to SSP engine
}

//-----------------------------------------------------------------------------
// Read 32 Bit OTP register. 
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		MOSI == 0
//		SCK == 1
// 		SPI-Mode == 1
//-----------------------------------------------------------------------------
static void Sensor_ReadOTP32 (DWORD *dwOTP)
{
	BYTE bShift;
	char i;
	short b, bit;

	// -- activate SPI mode (default is I2C)
	Goto_SPI(); 

	// -- set nCS signal to low (this will unwantedly be interpreted as "start normal read" by the sensor)
	//    however, this condition will be cleared when nCS is pulled back to high
	GPIO_LOW(nCS)

	// Set SCL signal to Low
	I2CHWSCL_GlobalSelect_ADDR &= ~I2CHWSCL_MASK; // take manual control
	GPIO_LOW(I2CHWSCL)
	
	// -- set MOSI signal to high
	MOSI_GlobalSelect_ADDR &= ~MOSI_MASK;	// take manual control
	GPIO_HIGH(MOSI)

	// -- positive pulse on nCS (2us setup time, 3us hold time)
	DELAY_8CYCLES
	GPIO_HIGH(nCS)
	DELAY_8CYCLES		
	GPIO_LOW(nCS)
	DELAY_16CYCLES
		
	// -- positive pulse on SCL
	GPIO_HIGH(I2CHWSCL)
	DELAY_8CYCLES
	GPIO_LOW(I2CHWSCL)
	DELAY_8CYCLES
	
	// -- configure MOSI to be an input
	GPIO_DM_HIGHZ(MOSI)
	
	// -- finally nCS high again to start read mode
	GPIO_HIGH(nCS)
	
	// -- read 4 Bytes
	for (b=0; b<=3; b++)	// 4 bytes to read
	{
		// -- read 8 bits
		for (bit=7; bit>=0; bit--)
		{
			// read one bit
			bShift <<= 1;
			if (MOSI_Data_ADDR & MOSI_MASK)
				bShift++;
			// positive clock pulse
			GPIO_HIGH(I2CHWSCL)
			GPIO_LOW (I2CHWSCL)
		}
		( (BYTE*)dwOTP )[b] = bShift;
	}
	
	// -- AS504x stops driving the MOSI signal
	GPIO_LOW(nCS)

	// -- Drive MOSI line as output (strong slow)
	GPIO_LOW(MOSI)
	GPIO_DM_STRONGSLOW(MOSI)
		
	GPIO_HIGH(nCS)	// stop OTP read

	// -- give control of the MOSI and SCK lines back to the SPI hardware.
	Goto_I2C();
}

//-----------------------------------------------------------------------------
// Write 32 Bit OTP register. 
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		MOSI == 0
//		SCK == 1
// 		SPI-Mode == 1
//-----------------------------------------------------------------------------
static void Sensor_WriteOTP32 (DWORD dwOTP)
{
	BYTE	bShift;
	short	i;
	BYTE	Status;
	
	// take manual control of MOSI and SCK
	Goto_SPI();

	// change SPI clock mode
	SPIM_Stop();
	SPIM_Start(SPIM_SPIM_MODE_0 | SPIM_SPIM_MSB_FIRST);
	
	// -- set nCS signal to low (this will unwantedly be interpreted as "start normal read" by the sensor)
	//    however, this condition will be cleared when nCS is pulled back to high
	GPIO_LOW(nCS)
	
	// -- set MOSI signal to high
	MOSI_GlobalSelect_ADDR &= ~MOSI_MASK;	// take manual control of the signal
	GPIO_HIGH(MOSI)
		
	// -- set nCS signal back to high (2us setup and hold time)
	DELAY_8CYCLES
	GPIO_HIGH(nCS)
	DELAY_8CYCLES

	// -- give control of the MOSI line back to the SPI hardware.
	MOSI_GlobalSelect_ADDR |= MOSI_MASK; 

	
	while( !( SPIM_bReadStatus() & SPIM_SPIM_TX_BUFFER_EMPTY ) );   	
	// -- tramsmit data
	for(i=0; i<=3; i++)
	{
		bShift = ( (BYTE*)&dwOTP )[i];
		SPIM_SendTxData( bShift );
		while( !( SPIM_bReadStatus() & SPIM_SPIM_SPI_COMPLETE ) );   	
	}
	
	// -- negative pulse on nCS 
	GPIO_LOW(nCS)
	GPIO_HIGH(nCS)

	// -- restore SPI clock mode
	SPIM_Start(SPIM_SPIM_MODE_1 | SPIM_SPIM_MSB_FIRST);
	// -- give control of the MOSI and SCK lines back to the SPI hardware.
	Goto_I2C();
}

//-----------------------------------------------------------------------------
// Set 16 Bit OTP register. Reads the current (32 Bit) OTP register, and modifies
// the user editable 16 bits of it.
//
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		MOSI == 0
//		SCK == 1
// 		SPI-Mode == 1
//-----------------------------------------------------------------------------
static void Sensor_SetOTP (WORD wOTP)
{
	DWORD dwOTP;
	Sensor_ReadOTP32 (&dwOTP);
	( (WORD*)&dwOTP )[1] = wOTP;
	Sensor_WriteOTP32(dwOTP);
}

//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// cannot be handled with the standard functionality in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//
//	return Value: number of Parameter bytes (1-31) that are 
//				  returned in the IN report
//-----------------------------------------------------------------------------
BYTE Sensor_Specific_Feature(void)
{
	BYTE ReadMode;
	union 
	{
		DWORD dw;
		BYTE  b[4];
	} Data;
	
	
	if (FEATURE_RPT.OUT.Target == eSensor)
	{
		switch (FEATURE_RPT.OUT.ParamSize)
		{
		case 4:
			// --- Read SIU		
			if ( (FEATURE_RPT.OUT.Get_nSet) &&
				 (FEATURE_RPT.OUT.Index_LSB == TYPE_SIU|ADDR_SIU) )
			{   // GET
				Get_Param_8(PAR8_READMODE, VOLATILE, &ReadMode);
				if ((ReadMode &0x02) != 0)
					AlignMode_Enter();
				Data.dw = I2C_Read_32bit(TYPE_SIU|ADDR_SIU);
				if ((ReadMode &0x02) != 0)
					AlignMode_Exit();
					
				FEATURE_RPT.IN.Param[0] = Data.b[0];
				FEATURE_RPT.IN.Param[1] = Data.b[1];
				FEATURE_RPT.IN.Param[2] = Data.b[2];
				FEATURE_RPT.IN.Param[3] = Data.b[3];
				return 4;
			}
			break;
		
		case 3:
			// --- Read or Write SINCOS signal bus
			if ( FEATURE_RPT.OUT.Index_LSB == TYPE_ADC|ADDR_ADC ) 
			{
				 if (FEATURE_RPT.OUT.Get_nSet)
				 {  // GET
					Data.dw = I2C_Read_24bit(TYPE_ADC|ADDR_ADC);
					FEATURE_RPT.IN.Param[0] = Data.b[1];
					FEATURE_RPT.IN.Param[1] = Data.b[2];
					FEATURE_RPT.IN.Param[2] = Data.b[3];
					return 3;
				 }
				 else
				 {  // SET
				 	Data.b[0] = 0;
				 	Data.b[1] = FEATURE_RPT.OUT.Param[0];
				 	Data.b[2] = FEATURE_RPT.OUT.Param[1];
				 	Data.b[3] = FEATURE_RPT.OUT.Param[2];
				 	I2C_Write_24bit(TYPE_ADC|ADDR_ADC, Data.dw);
				 	return 0;
				 }
			}
			break;
			 	
		case 1:
			// --- Read or Write AGC
			if ( FEATURE_RPT.OUT.Index_LSB == TYPE_AGC|ADDR_AGC )
			{
				if (FEATURE_RPT.OUT.Get_nSet)
				{ // GET
					FEATURE_RPT.IN.Param[0] = I2C_Read_8bit(FEATURE_RPT.OUT.Index_LSB);
					return 1;
				}
				else
				{ // SET
				 	I2C_Write_8bit(FEATURE_RPT.OUT.Index_LSB, FEATURE_RPT.OUT.Param[0]);
				 	return 0;
				}
			}
			
			if   ( (FEATURE_RPT.OUT.Index_LSB & TYPE_MASK) == TYPE_HALLFE) 
			{
				if (FEATURE_RPT.OUT.Get_nSet)
				{ // GET
					FEATURE_RPT.IN.Param[0] = I2C_Read_8bit(FEATURE_RPT.OUT.Index_LSB);
					return 1;
				}
				else
				{ // SET
				 	I2C_Write_8bit(FEATURE_RPT.OUT.Index_LSB, FEATURE_RPT.OUT.Param[0]);
				 	return 0;
				}
			}
			break;
		}
	}
	return 0;
}

//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// sets / gets a 8/16/32/168bit parameter in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//-----------------------------------------------------------------------------
void Sensor_Specific_Parameter(BYTE *NumberOfReturnedBytes)
{
	// Param_16[3] is to set the sensor's OTP register (user part)
	if ( (FEATURE_RPT.OUT.Index_LSB == 3) &&
		 (FEATURE_RPT.OUT.ParamSize == 2) &&
		 (FEATURE_RPT.OUT.Get_nSet  == 0) )
		SetOTP = TRUE;
		
	// Param_16[4] is to set the sensor's OTP register (factory settings part)
	if ( (FEATURE_RPT.OUT.Index_LSB == 4) &&
		 (FEATURE_RPT.OUT.ParamSize == 2) &&
		 (FEATURE_RPT.OUT.Get_nSet  == 0) )
		SetOTP32 = TRUE;

	// Param_8[3] is the Readmode, and requires to power down the sensor shortly
	if ( (FEATURE_RPT.OUT.Index_LSB == 3) &&
		 (FEATURE_RPT.OUT.ParamSize == 1) &&
		 (FEATURE_RPT.OUT.Get_nSet  == 0) )
		FSM = ePowerdown;
		
}

//-----------------------------------------------------------------------------
// void Sensor_FSM_Measure(void)
//
//	finite state machine to control the measurement process for 
//  the Sensor
//
//  IMPORTANT: 	call this routine once every millisecond. Otherwise the timing
//				parameters cannot be met.
//-----------------------------------------------------------------------------
void Sensor_FSM_Measure() 
{
	static WORD AngleRaw, Angle; 
	static BYTE Magnitude;
	static BYTE Status; 
	WORD   Value;
	BYTE   ReadMode;
	union
	{
		DWORD dw;
		BYTE   b[4];
	}	Val;

	// decrement millisecond counter
	if (Countdown_1ms != 0)
		Countdown_1ms--;

	// process actual state machine
	switch (FSM)
	{
		case ePowerdown:
			Sensor_Suspend();	// switch off power from the sensor
			Countdown_1ms = 200; // keep Sensor in the OFF state for 200ms

			FSM = ePowerup;
			break;
		
		case ePowerup:
			if (Countdown_1ms == 0)
			{
				Sensor_Resume();
				FSM = eReset;
			}
			break;
			
		case eReset:
			Countdown_1ms = 3; // 3ms delay until first measurement finished

			SetOTP = TRUE;
			SetOTP32 = TRUE;
			FSM = eIdle;
			break;
			
		case eIdle: 
			if (Countdown_1ms == 0)
			{
				Get_Param_16(PAR16_SAMPLERATE_MS, VOLATILE, &Countdown_1ms);;
				FSM = eReadADC;					  
			}
					// write OTP register if it has changed
			if (SetOTP)
			{
				WORD OTP;
				Get_Param_16(PAR16_SENSOR_OTP, VOLATILE, &OTP);
				Sensor_SetOTP (OTP);		
				SetOTP = FALSE;
			}
			if (SetOTP32)
			{
				WORD dwOTP[2];
				Get_Param_16(PAR16_SENSOR_OTP       , VOLATILE, &dwOTP[1]);
				Get_Param_16(PAR16_SENSOR_OTPFACTORY, VOLATILE, &dwOTP[0]);
				Sensor_WriteOTP32( *((DWORD*)dwOTP) );
				SetOTP32 = FALSE;
			}
			break;			
			
		case eReadADC:
			// Angle or Align data?
			Get_Param_8(PAR8_READMODE, VOLATILE, &ReadMode);
			if ((ReadMode &0x02) == 0)
				Sensor_ReadSIU (&AngleRaw, &Magnitude, &Status);
			else
			{
				AlignMode_Enter();
				Sensor_ReadSIU (&AngleRaw, &Magnitude, &Status);
				AlignMode_Exit();
			}
			Angle = Ang_MSB[AngleRaw / 0x100] + Ang_LSB[AngleRaw & 0xFF];
			Report.Ch1_LSB = Angle & 0xFF;
			Report.Ch1_MSB = Angle / 0x100;

			Report.Ch2_LSB = Magnitude;
			Report.Ch2_MSB = 0x00;

			Report.Ch3_LSB = Status;
			Report.Ch3_MSB = 0x00;

			NewSample = TRUE;

			FSM = eIdle;
			break;			
	}
}

//-----------------------------------------------------------------------------
void Sensor_Debug(void)
{
	/* currently no debug function */
}